# -*- coding: utf-8 -*-

import six

from subprocess import call
from subprocess import check_output
from subprocess import STDOUT


class Shell(object):

    @staticmethod
    def _run(method, cmd, cwd=None, shell=True, universal_newlines=True,
             stderr=STDOUT):
        """Internal wrapper for `call` amd `check_output`"""
        if not cmd:
            error_msg = 'Passed empty text or list'
            raise AttributeError(error_msg)
        if isinstance(cmd, six.string_types):
            cmd = str(cmd)
        if shell:
            if isinstance(cmd, list):
                cmd = ' '.join(cmd)
        else:
            if isinstance(cmd, str):
                cmd = cmd.strip().split()
        out = method(cmd, shell=shell, cwd=cwd, stderr=stderr,
                     universal_newlines=universal_newlines)
        if isinstance(out, bytes):
            out = out.decode('utf-8')
        return str(out).strip()

    @staticmethod
    def call(cmd, shell=True, cwd=None, universal_newlines=True, stderr=STDOUT):
        """Just execute a specific command."""
        return Shell._run(call, cmd, shell=shell, cwd=cwd, stderr=stderr,
                          universal_newlines=universal_newlines)

    @staticmethod
    def check_output(cmd, shell=True, cwd=None, universal_newlines=True,
                     stderr=STDOUT):
        """Execute a specific command and return the output."""
        return Shell._run(check_output, cmd, shell=shell, cwd=cwd,
                          stderr=stderr, universal_newlines=universal_newlines)



# -*- coding: utf-8 -*-

import sys

from argparse import ArgumentParser
from argparse import RawTextHelpFormatter
from argparse import SUPPRESS
from argparse import _HelpAction

from os import sep
from os.path import isdir
from os.path import isfile

from sklearn.externals import joblib

from sklearn_porter import Porter
from sklearn_porter import meta
from sklearn_porter.language import *


def parse_args(args):
    version = meta.get('version')
    header = '''
             #
### ### ### ### ### ### 
# # # # #    #  ##  #
### ### #    ## ### #   v{}
#'''.format(version)

    summary = dict(
        usage=header,
        description=meta.get('description'),
        epilog='More details on ' + meta.get('url'),
        formatter_class=RawTextHelpFormatter,
        add_help=False
    )
    p = ArgumentParser(**summary)

    # Remove the default arguments group:
    p._action_groups.pop()

    # File arguments:
    files = p.add_argument_group('File arguments')
    help = 'Path to an exported estimator in pickle (.pkl) format.'
    files.add_argument('input', help=help)
    help = 'Path to the output directory where ' \
           'the transpiled estimator will be stored.'
    files.add_argument('--to', required=False, help=help)

    # Template arguments:
    templates = p.add_argument_group('Template arguments')
    templates.add_argument('--class_name', default=None, required=False,
                           help='Define a custom class name.')
    templates.add_argument('--method_name', default='predict', required=False,
                           help='Define a custom method name.')

    # Optional arguments:
    optional = p.add_argument_group('Optional arguments')
    optional.add_argument('--export', '-e', required=False, default=False,
                          action='store_true', help='Whether to export '
                                                    'the model data or not.')
    optional.add_argument('--checksum', '-s', required=False, default=False,
                          action='store_true', help='Whether to append the '
                                                    'checksum to the filename '
                                                    'or not.')
    optional.add_argument('--data', '-d', required=False, default=False,
                          action='store_true', help='Whether to export just '
                                                    'the model data or all.')
    optional.add_argument('--pipe', '-p', required=False, default=False,
                          action='store_true', help='Print the transpiled '
                                                    'estimator to the console.')

    # Programming languages:
    langs = p.add_argument_group('Programming languages')
    languages = {key: clazz.LABEL for key, clazz in list(LANGUAGES.items())}
    langs.add_argument('--language', '-l', choices=languages.keys(),
                       default='java', required=False, help=SUPPRESS)
    for key, label in list(languages.items()):
        help = 'Set \'{}\' as the target programming language.'.format(label)
        langs.add_argument('--{}'.format(key), action='store_true', help=help)

    # Extra arguments:
    extras = p.add_argument_group('Extra arguments')
    extras.add_argument('--version', '-v', action='version',
                        version='sklearn-porter v{}'.format(version),
                        help='Show the version number and exit.')
    extras.add_argument('--help', '-h', action=_HelpAction,
                        help="Show this help message and exit.")

    # Show help by default:
    if len(sys.argv) == 1:
        p.print_help(sys.stderr)
        sys.exit(1)

    # Return dictionary:
    args = vars(p.parse_args(args))
    return args


def main():
    args = parse_args(sys.argv[1:])

    # Check input data:
    pkl_file_path = str(args.get('input'))
    if not isfile(pkl_file_path):
        exit_msg = 'No valid estimator in pickle ' \
                   'format was found at \'{}\'.'.format(pkl_file_path)
        sys.exit('Error: {}'.format(exit_msg))

    # Load data:
    estimator = joblib.load(pkl_file_path)

    # Determine the target programming language:
    language = str(args.get('language'))  # with default language
    languages = ['c', 'java', 'js', 'go', 'php', 'ruby']
    for key in languages:
        if args.get(key):  # found explicit assignment
            language = key
            break

    # Define destination path:
    dest_dir = str(args.get('to'))
    if dest_dir == '' or not isdir(dest_dir):
        dest_dir = pkl_file_path.split(sep)
        del dest_dir[-1]
        dest_dir = sep.join(dest_dir)

    # Port estimator:
    try:
        class_name = args.get('class_name')
        method_name = args.get('method_name')
        with_export = bool(args.get('export'))
        with_checksum = bool(args.get('checksum'))
        porter = Porter(estimator, language=language)
        output = porter.export(class_name=class_name, method_name=method_name,
                               export_dir=dest_dir, export_data=with_export,
                               export_append_checksum=with_checksum,
                               details=True)
    except Exception as exception:
        # Catch any exception and exit the process:
        sys.exit('Error: {}'.format(str(exception)))
    else:
        # Print transpiled estimator to the console:
        if bool(args.get('pipe', False)):
            print(output.get('estimator'))
            sys.exit(0)

        only_data = bool(args.get('data'))
        if not only_data:
            filename = output.get('filename')
            dest_path = dest_dir + sep + filename
            # Save transpiled estimator:
            with open(dest_path, 'w') as file_:
                file_.write(output.get('estimator'))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-


class JavaScript(object):

    KEY = 'js'
    LABEL = 'JavaScript'

    DEPENDENCIES = ['node']
    TEMP_DIR = 'js'
    SUFFIX = '.js'

    CMD_COMPILE = None

    # node estimator.js <args>
    CMD_EXECUTE = 'node {dest_dir}/{dest_file}'



# -*- coding: utf-8 -*-


class Go(object):

    KEY = 'go'
    LABEL = 'Go'

    DEPENDENCIES = ['go']
    TEMP_DIR = 'go'
    SUFFIX = '.go'

    # go build -o tmp/estimator tmp/estimator.go
    CMD_COMPILE = 'go build -o {dest_dir}/{dest_file} {src_dir}/{src_file}'

    # tmp/estimator <args>
    CMD_EXECUTE = '{dest_dir}/{dest_file}'



# -*- coding: utf-8 -*-

from sklearn_porter.language.C import C
from sklearn_porter.language.Go import Go
from sklearn_porter.language.Java import Java
from sklearn_porter.language.JavaScript import JavaScript
from sklearn_porter.language.PHP import PHP
from sklearn_porter.language.Ruby import Ruby

LANGUAGES = {
    C.KEY: C,
    Go.KEY: Go,
    Java.KEY: Java,
    JavaScript.KEY: JavaScript,
    PHP.KEY: PHP,
    Ruby.KEY: Ruby
}

__all__ = ['C', 'Go', 'Java', 'JavaScript', 'PHP', 'Ruby', 'LANGUAGES']



# -*- coding: utf-8 -*-


class Java(object):

    KEY = 'java'
    LABEL = 'Java'

    DEPENDENCIES = ['java', 'javac']
    TEMP_DIR = 'java'
    SUFFIX = '.java'

    # javac {class_path} tmp/Estimator.java
    # class_path = '-cp ./gson.jar'
    CMD_COMPILE = 'javac {class_path} {src_dir}/{src_file}'

    # java {class_path} Estimator <args>
    # class_path = '-cp ./gson.jar:./tmp'
    CMD_EXECUTE = 'java {class_path} {dest_dir}/{dest_file}'



# -*- coding: utf-8 -*-

import os
import sys

try:
    from shutil import which
except ImportError:
    def which(cmd, mode=os.F_OK | os.X_OK, path=None):
        """Given a command, mode, and a PATH string, return the path which
        conforms to the given mode on the PATH, or None if there is no such
        file.
        `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
        of os.environ.get("PATH"), or can be overridden with a custom search
        path.
        Note: This function was backported from the Python 3 source code.

        # __author__ = "Daniel Roy Greenfeld"
        # __email__ = "pydanny@gmail.com"
        # __version__ = "0.5.3"
        # https://github.com/pydanny/whichcraft/blob/master/whichcraft.py
        """
        def _access_check(fn, mode):
            return os.path.exists(fn) and os.access(fn, mode)\
                   and not os.path.isdir(fn)
        if os.path.dirname(cmd):
            if _access_check(cmd, mode):
                return cmd
            return None
        if path is None:
            path = os.environ.get('PATH', os.defpath)
        if not path:
            return None
        path = path.split(os.pathsep)
        if sys.platform == 'win32':
            if os.curdir not in path:
                path.insert(0, os.curdir)
            pathext = os.environ.get('PATHEXT', '').split(os.pathsep)
            if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
                files = [cmd]
            else:
                files = [cmd + ext for ext in pathext]
        else:
            files = [cmd]
        seen = set()
        for dir in path:
            normdir = os.path.normcase(dir)
            if normdir not in seen:
                seen.add(normdir)
                for thefile in files:
                    name = os.path.join(dir, thefile)
                    if _access_check(name, mode):
                        return name
        return None


class Environment(object):

    @staticmethod
    def read_sklearn_version():
        """Determine the installed version of sklearn"""
        from sklearn import __version__ as sklearn_ver
        sklearn_ver = str(sklearn_ver).split('.')
        sklearn_ver = [int(v) for v in sklearn_ver]
        major, minor = sklearn_ver[0], sklearn_ver[1]
        patch = sklearn_ver[2] if len(sklearn_ver) >= 3 else 0
        return major, minor, patch

    @staticmethod
    def _platform_is_windows(platform=sys.platform):
        """Is the current OS a Windows?"""
        matched = platform in ('cygwin', 'win32', 'win64')
        if matched:
            error_msg = "Windows isn't supported yet"
            raise OSError(error_msg)
        return matched

    @staticmethod
    def has_app(name, check_platform=True):
        """Check whether the application <name> is installed."""
        if check_platform:
            Environment._platform_is_windows()
        return which(str(name)) is not None

    @staticmethod
    def has_apps(names, check_platform=True):
        """Check whether the applications [<name>, ...] are installed."""
        if check_platform:
            Environment._platform_is_windows()
        for name in names:
            yield Environment.has_app(str(name), check_platform=False)

    @staticmethod
    def check_deps(deps):
        """check whether specific requirements are available."""
        if not isinstance(deps, list):
            deps = [deps]
        checks = list(Environment.has_apps(deps))
        if not all(checks):
            for name, available in list(dict(zip(deps, checks)).items()):
                if not available:
                    error_msg = "The required application/dependency '{0}'" \
                                " isn't available.".format(name)
                    raise SystemError(error_msg)



# -*- coding: utf-8 -*-

import logging


LOGGING_FORMAT = '%(name)-12s: %(levelname)-8s %(message)s'


class Logging(object):

    @staticmethod
    def get_logger(name, level=0):
        """Setup a logging instance"""
        level = 0 if not isinstance(level, int) else level
        level = 0 if level < 0 else level
        level = 4 if level > 4 else level
        console = logging.StreamHandler()
        level = [logging.NOTSET, logging.ERROR, logging.WARN, logging.INFO,
                 logging.DEBUG][level]
        console.setLevel(level)
        formatter = logging.Formatter(LOGGING_FORMAT)
        console.setFormatter(formatter)
        logging.getLogger('').addHandler(console)
        return logging.getLogger(name)



# -*- coding: utf-8 -*-


class C(object):

    KEY = 'c'
    LABEL = 'C'

    DEPENDENCIES = ['gcc']
    TEMP_DIR = 'c'
    SUFFIX = '.c'

    # gcc tmp/estimator.c -std=c99 -lm -o tmp/estimator
    CMD_COMPILE = 'gcc {src_dir}/{src_file} -std=c99 -lm -o {dest_dir}/{dest_file}'

    # tmp/estimator <args>
    CMD_EXECUTE = '{dest_dir}/{dest_file}'



# -*- coding: utf-8 -*-


class PHP(object):

    KEY = 'php'
    LABEL = 'PHP'

    DEPENDENCIES = ['php']
    TEMP_DIR = 'php'
    SUFFIX = '.php'

    CMD_COMPILE = None

    # php -f {} Estimator.php <args>
    CMD_EXECUTE = 'php -f {dest_dir}/{dest_file}'



# -*- coding: utf-8 -*-

from os.path import abspath
from os.path import dirname
from os.path import exists
from os.path import join
from json import load

from sklearn_porter.Porter import Porter


def _load_meta(path):
    """
    Load meta data about this package from file pypi.json.
    :param path: The path to pypi.json
    :return: Dictionary of key value pairs.
    """
    with open(path) as f:
        meta = load(f, encoding='utf-8')
        meta = {k: v.decode('utf-8') if isinstance(v, bytes) else v
                for k, v in meta.items()}

        src_dir = abspath(dirname(path))

        if 'requirements' in meta and \
                str(meta['requirements']).startswith('file://'):
            req_path = str(meta['requirements'])[7:]
            req_path = join(src_dir, req_path)
            if exists(req_path):
                reqs = open(req_path, 'r').read().strip().split('\n')
                reqs = [req.strip() for req in reqs if 'git+' not in req]
                meta['requirements'] = reqs
            else:
                meta['requirements'] = ''

        if 'long_description' in meta and \
                str(meta['long_description']).startswith('file://'):
            readme_path = str(meta['long_description'])[7:]
            readme_path = join(src_dir, readme_path)
            if exists(readme_path):
                readme = open(readme_path, 'r').read().strip()
                meta['long_description'] = readme
            else:
                meta['long_description'] = ''

    return meta


package = join(abspath(dirname(__file__)), 'pypi.json')
meta = _load_meta(package)

__author__ = meta.get('author')
__email__ = meta.get('author_email')
__license__ = meta.get('license')
__version__ = meta.get('version')



# -*- coding: utf-8 -*-


class Ruby(object):

    KEY = 'ruby'
    LABEL = 'Ruby'

    DEPENDENCIES = ['ruby']
    TEMP_DIR = 'ruby'
    SUFFIX = '.rb'

    CMD_COMPILE = None

    # ruby estimator.rb <args>
    CMD_EXECUTE = 'ruby {dest_dir}/{dest_file}'
